module Main where

import System (getArgs,getProgName,exitFailure,ExitCode (..),exitWith)
import CliWrap (Cli (..), toCliWithLog, cliDoVerbose, cliDoLog, getCliFiles, getCliLogfile, getCliStartStop)
import IPC (toDNAMmodelSpec,toDNAMmodelSpecWithLog,DNAMmodelSpec(..))
import LogResult (newLR, closeLogLR, getResLR, LogResult)
import Utils ((?),numberLines,isNothing,fromJust)

ipcVersion = "0.91a 230603 (\"James Clerk Maxwell\" Release)"
ipcMessage = unlines ["** The Imperial PEPA compiler:",
		      "ipc v"++ipcVersion,
		      "DNAmaca generator: Jeremy Bradley; PEPA parser: Stephen Gilmore",
		      "Incorporating Sven Panne's port of the GNU GetOpt library",
		      "Produced on location in/between Edinbugh, Leeds, London and Ludlow, 2002-3",
		      "Mail queries and problems to: jb@doc.ic.ac.uk"]

main :: IO()
main = do {
		args <- getArgs; name <- getProgName;
		let 
		  argsLog = newLR "Main" analyseArgs args
		  cliLog = toCliWithLog ipcVersion name argsLog
		  (log, opts) = closeLogLR ipcMessage cliLog
		  (infile, outfile) = getCliFiles opts
		  logfile = getCliLogfile opts
		  actionLbls = getCliStartStop opts
		  prelogfile = "ipc.log"
		in case opts of
		  CliValid os a    -> do {
					  s <- readFile infile;
					  let
					    probename = "probe"
					    def = (probename,actionLbls)
			  		    finalOut = closeLogLR log (toDNAMmodelSpecWithLog def s)
					    mLog = (cliDoLog opts) ? (Just logfile, Nothing)
					    mess = (cliDoVerbose opts) ? (fst finalOut, ipcMessage)
					    process = concat ["Compiling: ", infile, " -> ", outfile]
					  in do {
						 putStrLn (concat [mess, "\n", process]);
						 writeLogResults (mLog,outfile) ipcMessage finalOut;
						 exitWith ExitSuccess; 
						 }
					 }
		  CliInfo  _ _ inf -> if (cliDoLog opts)
		                      then do { writeFile prelogfile log; putStrLn inf; exitWith ExitSuccess; }
		                      else do { putStrLn inf; exitWith ExitSuccess; } 
		  CliError _ _ err -> if (cliDoLog opts)
		                      then do { writeFile prelogfile log; putStrLn err; exitFailure; }
		                      else do { putStrLn err; exitFailure; } 
	  }


writeLogResults :: Show a => (Maybe String,String) -> String -> (String,a) -> IO ()
writeLogResults (mfnl,fnr) ver (log,res)
    | isNothing mfnl = writeFile fnr res'
    | otherwise = ((writeFile (fromJust mfnl) log) >> (writeFile fnr res')) where
                     res' = ((unlines (map ("%% "++) $ lines ver)) ++ (show res))

analyseArgs :: [String] -> String
analyseArgs as = concat [show $ length as, " command line arguments(s) extracted:\n", numberLines as]

