module PepaWrap (
		 toPEPAsystem, toPEPAmodel, toPEPAmodelWithLog, toPEPAsystemWithLog, 
		 toGPS, toGPSwithLog, toGPSrtnWithLog, mkPEPAtransList, getPEPApCpts,
		 nextPEPAcpt, fromPEPArate, isRateSymbol, mkLbl, processPEPArate, toGPSrtn,
		 PEPAsystem, PEPAmodel (..), PEPAact, PEPArate, PEPAcpt, PEPAprobeDef,
		 PEPAtrans, PEPAdb, PEPAtransList, IPCsystemRtn (..),
		 GenericParallelStructure (..), GPSelem (..)
		) where

import PepaParser (Component(..), parsePepa, parsePepaWithLog)
import PepaTypes (Identifier, PEPArate, Rate, processPEPArate, fromPEPArate, isRateSymbol, mkLbl)
import Utils (fromJust,isNothing,isRight,isLeft,fromRight,fromLeft,selectTerm,indent,numberLines,(?))
import LogResult (LogResult, mapLR, closeLogLR)
import List (find,(\\),nub)

data IPCsystemRtn a  = IPCsr a (Identifier, Identifier)
type PEPAprobeDef    = (Identifier, ([Identifier], [Identifier])) -- probe label, (start actions, stop actions)
type PEPAsystem      = IPCsystemRtn PEPAmodel

data PEPAmodel = PEPAmod PEPAcpt PEPAdb
type PEPAdb = [(PEPAcpt, Either PEPAp PEPAs)] 
data PEPAp = PEPAcoop PEPAcpt PEPAcpt [PEPAact] | PEPAhide PEPAcpt [PEPAact] deriving Show
data PEPAs = PEPAsum PEPAcpt PEPAcpt | PEPAprefix PEPAtrans PEPAcpt deriving Show
type PEPAcptDb = [(PEPAcpt, Int)]
type PEPAtrans = (PEPAact, PEPArate)
type PEPAtransList = [(PEPAcpt, PEPAtrans, PEPAcpt)] 
type PEPAcpt = Identifier
type PEPAact = Identifier

data GPSelem a  => GenericParallelStructure a
    = PAR (GenericParallelStructure a) (GenericParallelStructure a) [PEPAact] | SEQ a

class GPSelem a where
    toGPSelem :: Int -> PEPAmodel -> a
    documentGPSelem :: a -> String

instance Show PEPAmodel where
    show (PEPAmod ptr db) = numberLines (ptr:map show db)

toPEPAsystem :: PEPAprobeDef -> String -> PEPAsystem
toPEPAsystem def = (mkPEPAsystem def) . parsePepa

toPEPAsystemWithLog :: PEPAprobeDef -> String -> LogResult PEPAsystem
toPEPAsystemWithLog def = (mapLR "PepaWrap [mkPEPAsystem]" (mkPEPAsystem def) (analysePEPAsystem def)) . toPEPAcomponentWithLog

toPEPAmodel :: String -> PEPAmodel
toPEPAmodel = convSTGpepaJTBpepa . parsePepa

toPEPAmodelWithLog :: String -> LogResult PEPAmodel
toPEPAmodelWithLog = (mapLR "PepaWrap" convSTGpepaJTBpepa analysePEPAmodel) . toPEPAcomponentWithLog

toPEPAcomponentWithLog :: String -> LogResult Component
toPEPAcomponentWithLog = (mapLR "PepaWrap [convSTGpepaJTBpepa]" id analysePEPAcomponent) . parsePepaWithLog

mkPEPAsystem :: PEPAprobeDef -> Component -> PEPAsystem
mkPEPAsystem pdf@(_,(ss,ts)) cpt
    | not.and $ map (\x -> elem x ats) (ss ++ ts) = error "PepaWrap.mkPEPAsystem: actions in probe specification do not exist in model"
    | (not.noPEPAduplicateConsts) model = error "PepaWrap.mkPEPAsystem: duplicate constant definitions discovered in model"
    | otherwise = (IPCsr model (mkStateLbl (id1, 0), mkStateLbl (id2, 0))) where
		  model = convSTGpepaJTBpepa cpt'
		  ats = getPEPAacts cpt
		  (IPCsr cpt' (id1,id2)) = addPEPAprobe pdf cpt

noPEPAduplicateConsts :: PEPAmodel -> Bool
noPEPAduplicateConsts (PEPAmod _ []) = True
noPEPAduplicateConsts (PEPAmod l1 (cp:cpts)) =  and (noPEPAduplicateConsts (PEPAmod l1 cpts) : map (\x -> fst cp /= fst x) cpts)

getPEPAacts :: Component -> [Identifier]
getPEPAacts (VAR _) = []
getPEPAacts (PREFIX ((act, _), c)) = (act : getPEPAacts c)
getPEPAacts (CHOICE (c1, c2)) = (getPEPAacts c1) ++ (getPEPAacts c2)
getPEPAacts (COOP (c1, c2, _)) = (getPEPAacts c1) ++ (getPEPAacts c2)
getPEPAacts (HIDING (c, _)) = getPEPAacts c
getPEPAacts (CONSTS (_, c1, c2)) = (getPEPAacts c1) ++ (getPEPAacts c2)


mkPEPAmodelPtrsReport :: PEPAmodel -> String
mkPEPAmodelPtrsReport mod = (null lst) ? ("...MODEL OK.", mess++show lst) where
			    lst = checkPEPAmodelPtrs mod 
			    mess = "Component labels for which no description exists (may cause later errors!):\n"

checkPEPAmodelPtrs :: PEPAmodel -> [PEPAcpt]
checkPEPAmodelPtrs mod@(PEPAmod _ cpts) = [c | c <- cps, (not.isPEPAcpt) (PEPAmod c cpts)] where 
					  cps = getPEPAcpts mod

getPEPAcpts :: PEPAmodel -> [PEPAcpt]
getPEPAcpts mod@(PEPAmod start cpts) = nub (start : (concat $ map getCpts cpts)) where
   getCpts :: (PEPAcpt, Either PEPAp PEPAs) -> [PEPAcpt]
   getCpts (_, Left (PEPAcoop c1 c2 _)) = [c1,c2]
   getCpts (_, Left (PEPAhide c _)) = [c]
   getCpts (_, Right (PEPAsum c1 c2)) = [c1,c2]
   getCpts (_, Right (PEPAprefix _ c)) = [c]

analysePEPAsystem :: PEPAprobeDef -> PEPAsystem -> String
analysePEPAsystem (lbl,(ss,ts)) (IPCsr mod (s1,s2)) = concat ["Create probe label: ", lbl, "\n"]
						       ++ numberLines ["Start actions: " ++ show ss,
								       "Stop actions: " ++ show ts,
								       "Probe stop state created: " ++ s1,
								       "Probe start state created: " ++ s2]
						       ++ (analysePEPAmodel mod)

analysePEPAmodel :: PEPAmodel -> String
analysePEPAmodel mod@(PEPAmod start cpts) 
    = unlines ["Root component label set to: " ++ start,
	      (show $ length cpts) ++ " component entries created in PEPA normal form database",
	      "PEPA normal form representation:", show mod,
	      "Checking consistency of modified PEPA model with probe description.",	   
	      mkPEPAmodelPtrsReport mod]

analysePEPAcomponent :: Component -> String
analysePEPAcomponent cpt = unlines ["New component labels have been created for PEPA normal form model",
			"to describe implicit states/structure in input model.",
			"Model normal form:",
			show $ convSTGpepaJTBpepa cpt,
    			"All state labels are listed below for each explicit PEPA component label:",
			(numberLines . lines . analyseCpt) cpt,
			"Checking consistency of unmodified PEPA model.",
			mkPEPAmodelPtrsReport $ convSTGpepaJTBpepa cpt] where 
    analyseCpt :: Component -> String			   
    analyseCpt (CONSTS (lbl1, cp, VAR lbl2)) = concat [lbl1, " -> ", getLbls lbl1 cp]
    analyseCpt (CONSTS (lbl, cp1, cp2@(CONSTS _))) = concat [lbl, " -> ", getLbls lbl cp1, "\n", analyseCpt cp2]
    analyseCpt (CONSTS (lbl, cp1, cp2)) = let nlbl = "0_root" in concat [lbl, " -> ", getLbls lbl cp1, "\n"] 
					  ++ concat ["Introduced dummy root label ", nlbl, "\n", nlbl, " -> ", getLbls nlbl cp2] 
    analyseCpt _ = error "analysePEPAcomponent: incomplete/corrupt PEPA specification"
    --
    getLbls :: Identifier -> Component -> String
    getLbls idr cp0 = foldr1 (\x -> \y -> x++", "++y) [ mkStateLbl (idr,i) | i <- [0..getPEPAdummyLblsInit idr cp0] ]

toGPSwithLog :: GPSelem a => LogResult PEPAmodel -> LogResult (GenericParallelStructure a)
toGPSwithLog = mapLR "PepaWrap [toGPS]" toGPS analyseGPS

analyseGPS :: GPSelem a => GenericParallelStructure a -> String
analyseGPS gps = concat [show cnt, " parallel component(s) identified\n", out] where
    (cnt, out) = analyse 0 gps
    analyse :: GPSelem a => Int -> GenericParallelStructure a -> (Int,String)
    analyse n (SEQ x) = (n+1, (documentGPSelem x)++"\n")
    analyse n (PAR x y as) = (m2, concat ["Coop set: ",show as,"\n",indent str]) where
			     (m1, lg1) = analyse n x; (m2, lg2) = analyse m1 y
			     str = concat ["--Cpt ",show m1,"--\n",lg1,"--Cpt ",show m2,"--\n",lg2]

toGPSrtnWithLog :: GPSelem a => LogResult PEPAsystem -> LogResult (IPCsystemRtn (GenericParallelStructure a))
toGPSrtnWithLog = mapLR "PepaWrap [toGPSrtn]" toGPSrtn analyseGPSrtn

analyseGPSrtn :: GPSelem a => IPCsystemRtn (GenericParallelStructure a) -> String
analyseGPSrtn (IPCsr gps is) = (analyseGPS gps) ++ "\nPEPA probe start/end states: " ++ (show is)

toGPSrtn :: GPSelem a => PEPAsystem -> IPCsystemRtn (GenericParallelStructure a)
toGPSrtn (IPCsr mod (i1,i2))
    | (isNothing f1 && isNothing f2) = error "PepaWrap.toGPSrtn: probe labels not found in parallel component list"
    | otherwise = (IPCsr (toGPS mod) is') where
		  (f1,f2) = (lookup i1 ixs, lookup i2 ixs)
		  is' = (isNothing f2) ? ((mkLbl (i1,fj1), mkLbl (i2,fj1)), (mkLbl (i1,fj2), mkLbl (i2,fj2)))
		  (fj1,fj2) = (fromJust f1, fromJust f2)
		  ixs = uniquePEPAcptIndex mod
			  

-- Creates a generic parallel structure of combined components
-- on the assumption that the new formalism has a conversion
-- to its own representation of sequential components
-- At the moment hiding is ignored.
toGPS :: GPSelem a => PEPAmodel -> GenericParallelStructure a
toGPS mod = snd $ mkGeneric (uniquePEPAcptIndex mod) mod where
    mkGeneric :: GPSelem a => PEPAcptDb -> PEPAmodel -> (PEPAcptDb, GenericParallelStructure a)
    mkGeneric ixs mod@(PEPAmod start cpts)
	| isRight ps  = (ixs', SEQ $ toGPSelem ix mod)
	| otherwise   = getCpts $ fromLeft ps where
	     ps = nextPEPAcpt mod
	     (ix, ixs') = newPEPAcptDb start ixs
	     getCpts :: GPSelem a => PEPAp -> (PEPAcptDb, GenericParallelStructure a)
	     getCpts (PEPAcoop c1 c2 acts) = (ixs2, PAR str1 str2 acts) where
		     (ixs1, str1) = mkGeneric ixs  (PEPAmod c1 cpts)
		     (ixs2, str2) = mkGeneric ixs1 (PEPAmod c2 cpts)
	     getCpts (PEPAhide c1 _) = mkGeneric ixs (PEPAmod c1 cpts) 

-- Allocates counters to each component label so that each instance in the PEPA parallel structure
-- can be given a unique reference in the DNAmaca description
newPEPAcptDb :: PEPAcpt -> PEPAcptDb -> (Int, PEPAcptDb)
newPEPAcptDb cp ixs 
    | isNothing fnd = (0, (cp,1):ixs)
    | otherwise     = (n, (cp,n+1):ixs')  where
	 fnd = selectTerm (\x -> (fst x) == cp) ixs
	 ((_,n),ixs') = fromJust fnd

-- Assigns unique index to all PEPA components
uniquePEPAcptIndex :: PEPAmodel -> [(PEPAcpt, Int)]
uniquePEPAcptIndex mod = zip (getPEPApCpts mod) [1..]
-- uniquePEPAcptIndex = uniqueIndex . getPEPApCpts where
--    uniqueIndex :: [PEPAcpt] -> [(PEPAcpt, Int)]
 --   uniqueIndex = (concat . map (flip zip [0..])) . (group.sort)


-- attempts to distinguish individual components
getPEPApCpts :: PEPAmodel -> [PEPAcpt]
getPEPApCpts mod@(PEPAmod start cpts)
    | isRight ps  = [start]
    | otherwise   = getCpts $ fromLeft ps where
	  ps = nextPEPAcpt mod
 	  getCpts :: PEPAp -> [PEPAcpt]
	  getCpts (PEPAcoop c1 c2 _) = (getPEPApCpts $ PEPAmod c1 cpts)
				      ++ (getPEPApCpts $ PEPAmod c2 cpts)
	  getCpts (PEPAhide c1 _) = (getPEPApCpts $ PEPAmod c1 cpts)

-- gets a list of possible transitions from a given start point in a sequential component
-- in the form: (state, transition, state)
mkPEPAtransList :: PEPAmodel -> PEPAtransList
mkPEPAtransList mod@(PEPAmod x _) = map (\(y,z) -> (x,y,z)) $ nextPEPAtransitions mod

-- gets a list of possible transitions from a given start point in a sequential component
-- in the form: (transition, state)
nextPEPAtransitions :: PEPAmodel -> [(PEPAtrans, PEPAcpt)]
nextPEPAtransitions mod@(PEPAmod start cpts)
    | isLeft ps   = error "nextPEPAstep: found parallel component in sequential defininition"
    | otherwise   = getTrans $ fromRight ps where
         ps = nextPEPAcpt mod
	 getTrans :: PEPAs -> [(PEPAtrans, PEPAcpt)]
	 getTrans (PEPAprefix tr c) = [(tr,c)]
	 getTrans (PEPAsum c1 c2) = (nextPEPAtransitions $ PEPAmod c1 cpts) 
				    ++ (nextPEPAtransitions $ PEPAmod c2 cpts)

-- looks up component definition in db
nextPEPAcpt :: PEPAmodel -> Either PEPAp PEPAs
nextPEPAcpt (PEPAmod start cpts)
    | isNothing qry  = error ("nextPEPAcpt: component "++start++" not found in model")
    | otherwise      = fromJust qry where
		       qry = lookup start cpts

isPEPAcpt :: PEPAmodel -> Bool
isPEPAcpt (PEPAmod start cpts) = (not.isNothing) (lookup start cpts)

convSTGpepaJTBpepa :: Component -> PEPAmodel
-- Terminating condition 1: where the root lbl2 is explicitly labelled
convSTGpepaJTBpepa (CONSTS (lbl1, cp, VAR lbl2)) = PEPAmod (mkStateLbl (lbl2, 0)) (getPEPAdbInit lbl1 cp)
-- Standard recursive case - where we are in the middle of a set of CONSTS definitions
convSTGpepaJTBpepa (CONSTS (lbl, cp1, cp2@(CONSTS _)))
    = let (PEPAmod lbl' db) = convSTGpepaJTBpepa cp2 in PEPAmod lbl' ((getPEPAdbInit lbl cp1) ++ db)
-- Terminating condition 2: where the root is not explicitly labelled and thus we have to provide a dummy label
-- Note that "0_root" is an illegal component name as far as the parser is concerned, so there is no danger of a label clash
convSTGpepaJTBpepa (CONSTS (lbl, cp1, cp2))
    = let nlbl = "0_root" in PEPAmod (mkStateLbl (nlbl, 0)) ((getPEPAdbInit lbl cp1) ++ (getPEPAdbInit nlbl cp2))
convSTGpepaJTBpepa _ = error "convSTGpepaJTBpepa: incomplete/corrupt PEPA specification"



-- Takes a single line of PEPA and returns number of explicit extra state labels required
getPEPAdummyLblsInit :: Identifier -> Component -> Int
getPEPAdummyLblsInit lbl cp = let (n,_,_) = getPEPAdb (lbl,0) cp in (n-1)

-- Takes a single line of PEPA and adds any explicit state labels as required
getPEPAdbInit :: Identifier -> Component -> PEPAdb
getPEPAdbInit lbl cp = getPEPAdbInit2 lbl 0 cp
-- Allows the setting of the index for a root component, useful for a dummy root variable
getPEPAdbInit2 :: Identifier -> Int -> Component -> PEPAdb
getPEPAdbInit2 lbl n cp = let (_,_,db) = getPEPAdb (lbl,n) cp in db

-- getPEPAdb: (base component label, dummy variable number) -> component
--    -> (component dummy variable index reached, pointer component name in db, db)
getPEPAdb :: (Identifier, Int) -> Component -> (Int, Identifier, PEPAdb)
getPEPAdb (lbl, i) cp
    = let lbl1 = mkStateLbl (lbl,i) in case cp of
      (CONSTS _)             -> error "getPEPAdb: should be no CONSTS in the definition"

      (PREFIX ((a,r),cp))    -> (n, lbl1, [(lbl1, Right $ PEPAprefix (a,processPEPArate r) lbl2)]++db) where
				 (n, lbl2, db) = getPEPAdb (lbl,i+1) cp

      (CHOICE (cp1, cp2))    -> (m, lbl1, [(lbl1, Right $ PEPAsum lbl2 lbl3)]++db2++db3) where
				 (n,lbl2,db2) = getPEPAdb (lbl,i+1) cp1
				 (m,lbl3,db3) = getPEPAdb (lbl,n) cp2

      (COOP (cp1,cp2, acts)) -> (m, lbl1, [(lbl1, Left $ PEPAcoop lbl2 lbl3 acts)]++db2++db3) where
				 (n,lbl2,db2) = getPEPAdb (lbl,i+1) cp1
				 (m,lbl3,db3) = getPEPAdb (lbl,n) cp2

      (HIDING (cp1, acts))   -> (n, lbl1, [(lbl1, Left $ PEPAhide lbl2 acts)]++db) where
				(n, lbl2, db) = getPEPAdb (lbl,i+1) cp1

      (VAR v)                -> (i,mkStateLbl (v,0),[])

addPEPAprobe :: PEPAprobeDef -> Component -> IPCsystemRtn Component
addPEPAprobe (lbl, (ss, ts)) sys = let (IPCsr probe lbls) = mkPEPAprobe (lbl, (ss, ts)) 
				       in (IPCsr (combineComponents ( ( ss \\ ts )++ ts) sys probe) lbls) where
    combineComponents :: [Identifier] -> Component -> Component -> Component
    combineComponents as pb (CONSTS (c1,c2,c3)) = CONSTS (c1,c2,combineComponents as pb c3)
      -- recurses down system component leaving it intact until in gets to final definition
    combineComponents as pb@(CONSTS _) syTerm = combineComponents as syTerm pb
      -- on finding system component defn, look at probe: if still not defn itself then switch with system defn
      -- so adding the probe CONST terms into the system component structure, as desired
    combineComponents as syTerm pbTerm = COOP (syTerm, pbTerm,as)
      -- when finally both are definitions (ie not CONSTS), we can compose both in a last COOP over the action set "as"
      -- Hmmm! :-)

mkPEPAprobe :: PEPAprobeDef -> IPCsystemRtn Component
mkPEPAprobe (lbl, (ss, ts)) = (IPCsr (CONSTS (thisLbl, outDesc, CONSTS (nextLbl, backDesc, VAR thisLbl))) (nextLbl,thisLbl)) where
		 thisLbl = "_"++lbl
		 nextLbl = thisLbl++"_run"
		 repRt = "infty"
		 outDesc = foldr1 (\x -> \y -> CHOICE (x,y)) 
			    ([PREFIX ((s, repRt), VAR nextLbl) | s <- ss] ++ [PREFIX ((t, repRt), VAR thisLbl) | t <- (ts\\ss)])
		 backDesc = foldr1 (\x -> \y -> CHOICE (x,y)) 
			    ([PREFIX ((s, repRt), VAR nextLbl) | s <- (ss\\ts)] ++ [PREFIX ((t, repRt), VAR thisLbl) | t <- ts])

mkStateLbl :: (Identifier,Int) -> Identifier
mkStateLbl = mkLbl

main :: String -> IO ()
main f1 = do {
	      s <- readFile f1;
	      let 
	        def  = ("probe",(["yes"],["help"]))
	        lg = fst $ closeLogLR "Test Pepa spec rig" (toPEPAsystemWithLog def s) 
	      in putStrLn lg
--	      let pdf = ("probe",(["yes"],["help"]))
--	          cpt = (parsePepa s) 
--	          (IPCsr mod lbls) = (mkPEPAsystem pdf cpt)
--	      in putStrLn $ concat [show lbls, "\n", show mod]
	     }

